from django.contrib import admin


class PaypalTransactionAdmin(admin.ModelAdmin):
    list_display = ('created_at', 'payment_id', 'reference_id', 'description',
                    'currency', 'amount_total', 'payer_email', 'state', 'is_succeed')
    search_fields = ('payment_id', 'invoice_number', 'payer_email')
    list_filter = ('state',)



from unittest.mock import Mock

from django.test import TestCase

from paypal.models import PaypalTransaction
from paypal.signals import payment_succeed


class PaypalTransactionTest(TestCase):
    def test_add_item(self):
        tx = PaypalTransaction()

        tx.add_item('simple', 'USD', 0.233)
        tx.add_item('hello', 'USD', 1.01,
                    quantity=3,
                    sku='HL',
                    description='test item',
                    tax=0.07)
        self.assertEqual(tx.currency, 'USD')
        self.assertEqual(str(tx.amount_total), '3.33')
        self.assertEqual(tx.amount_details['subtotal'], '3.26')
        self.assertEqual(tx.amount_details['tax'], '0.07')

    def test_signal_payment_succeed(self):
        tx = PaypalTransaction()
        tx.add_item('simple', 'USD', 0.233)
        tx.save()

        mock = Mock()
        payment_succeed.connect(mock)

        tx.state = PaypalTransaction.STATE_APPROVED
        tx.save()
        self.assertEqual(mock.call_count, 1)

        tx.save()
        self.assertEqual(mock.call_count, 1, '只有切换到APPROVED的时候才抛signal')



from django.conf.urls import url
from django.views.generic import TemplateView

from paypal.views import PaypalRedirectView, PaypalReturnView

urlpatterns = [
    url('^redirect/(?P<pk>.*?)/$', PaypalRedirectView.as_view(), name='paypal_redirect'),
    url('^return/$', PaypalReturnView.as_view(), name='paypal_return'),
    url('^cancel/$', TemplateView.as_view(template_name='paypal/cancel.html'), name='paypal_cancel'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-29 02:59
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields.json


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PaypalTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=127)),
                ('items', django_extensions.db.fields.json.JSONField(default=[])),
                ('reference_id', models.CharField(blank=True, max_length=256, null=True)),
                ('note_to_payer', models.CharField(blank=True, max_length=165, null=True)),
                ('custom', models.CharField(blank=True, max_length=127, null=True)),
                ('payment_id', models.CharField(blank=True, max_length=64, null=True)),
                ('invoice_number', models.CharField(blank=True, max_length=127, null=True)),
                ('amount_total', models.DecimalField(decimal_places=2, max_digits=10)),
                ('currency', models.CharField(choices=[('USD', 'USD')], max_length=3)),
                ('amount_details', django_extensions.db.fields.json.JSONField(default={})),
                ('payer_id', models.CharField(blank=True, max_length=16, null=True)),
                ('payer_email', models.EmailField(blank=True, max_length=254, null=True)),
                ('payer_data', django_extensions.db.fields.json.JSONField(default=dict, null=True)),
                ('payee_merchant_id', models.CharField(blank=True, max_length=16, null=True)),
                ('state', models.CharField(choices=[('created', 'created'), ('approved', 'approved'), ('failed', 'failed')], max_length=16)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-08 05:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paypal', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='paypaltransaction',
            name='state',
            field=models.CharField(choices=[('created', 'created'), ('approved', 'approved'), ('failed', 'failed')], default='created', max_length=16),
        ),
    ]



import random

from django.shortcuts import redirect, render
from django.views.decorators.csrf import csrf_exempt

from paypal.models import PaypalTransaction


@csrf_exempt
def sample_pay_view(request):
    if request.method == 'GET':
        return render(request, 'sample_pay.html')
    elif request.method == 'POST':
        tx = PaypalTransaction(
            description="sample pay",
        )
        tx.add_item("sample", 'USD', float(request.POST['amount']))
        tx.reference_id = str(random.randint(10000, 99999))
        tx.save()
        return redirect('paypal_redirect', pk=tx.pk)



from django.conf import settings
from django.db import models
from django.db.models.signals import post_save
from django_extensions.db.fields.json import JSONField
from model_utils import FieldTracker

from .signals import payment_succeed


# https://developer.paypal.com/docs/api/payments/#definition-transaction
class PaypalTransaction(models.Model):
    USD = 'USD'

    CURRENCIES = (
        (USD, USD),
    )

    STATE_CREATED = 'created'
    STATE_APPROVED = 'approved'
    STATE_FAILED = 'failed'

    STATES = (
        (STATE_CREATED, STATE_CREATED),
        (STATE_APPROVED, STATE_APPROVED),
        (STATE_FAILED, STATE_FAILED),
    )

    # request ==================================================================

    # The description of what is being paid for.
    description = models.CharField(max_length=127)

    # item_list.items: The item details.
    # sku[127], name[127], description[127], quantity[10], price[10], currency[3], tax, url
    items = JSONField(default=[])

    # Optional. The merchant-provided ID for the purchase unit.
    reference_id = models.CharField(max_length=256, null=True, blank=True)

    # A free-form field that clients can use to send a note to the payer.（non-transaction field)
    note_to_payer = models.CharField(max_length=165, null=True, blank=True)

    # A free-form field for clients' use.
    custom = models.CharField(max_length=127, null=True, blank=True)

    # paypal的支付id，需要在生成model后，调用api生成Payment对象，然后存入payment.id
    payment_id = models.CharField(max_length=64, null=True, blank=True)

    # The invoice number to track this payment.
    # 没有就不会显示在paypal中，可以用作商户订单号
    invoice_number = models.CharField(max_length=127, null=True, blank=True)

    # The amount to collect.
    # 除了details中的一些特殊的金额，其他都要手动设置，要通过add_item来更新
    amount_total = models.DecimalField(max_digits=10, decimal_places=2)
    currency = models.CharField(max_length=3, choices=CURRENCIES)
    amount_details = JSONField(default={})

    # SKIPPED
    # - note_to_payee: An optional note to the recipient of the funds in this transaction.
    # - purchase_order: purchase order is number or id specific to this payment
    # - soft_descriptor: The soft descriptor that is used when charging this funding source. If the string's length is greater than the maximum allowed length, the string is truncated.
    # - payment_options: The payment options requested for this transaction.
    # - notify_url: The URL to send payment notifications. （亲测可以收到通知）
    # - order_url: The URL on the merchant site related to this payment.
    # - related_resources: read only
    # - item_list
    #   + shipping_address
    #   + shipping_method
    #   + shipping_phone_number

    # response only ==================================================================

    # The source of the funds for this payment. Payment method is PayPal Wallet payment, bank direct debit, or direct credit card.
    # 订单成功后返回的数据中获得
    payer_id = models.CharField(max_length=16, null=True, blank=True)
    payer_email = models.EmailField(null=True, blank=True)
    payer_data = JSONField(null=True)

    # payee: The recipient of the funds in this transaction.
    payee_merchant_id = models.CharField(max_length=16, null=True, blank=True)

    # The state of the payment, authorization, or order transaction
    state = models.CharField(max_length=16, choices=STATES, default=STATE_CREATED)

    # customized ==================================================================
    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    tracker = FieldTracker(fields=['state'])

    def is_succeed(self):
        return self.state == self.STATE_APPROVED

    is_succeed.boolean = True

    def add_item(self, name, currency, price, quantity=1, sku=None, description=None, tax=None):
        item = {
            "name": name,
            "currency": currency,
            "price": "{:.2f}".format(price),
            "quantity": str(quantity),
        }

        if sku:
            item['sku'] = str(sku)

        if description:
            item['description'] = str(description)

        if tax:
            item['tax'] = "{:.2f}".format(tax)

        self.items.append(item)
        self.refresh_amount()

    def refresh_amount(self):
        currencies = set(x['currency'] for x in self.items)
        assert len(currencies) == 1, 'multi currency not supported'

        self.currency = list(currencies)[0]

        self.amount_details['subtotal'] = '{:.2f}'.format(sum(float(x['price']) * float(x['quantity'])
                                                              for x in self.items))

        tax = sum(float(x.get('tax', 0)) for x in self.items)
        if tax:
            self.amount_details['tax'] = '{:.2f}'.format(tax)
        else:
            self.amount_details.pop('tax', None)

        self.amount_total = float('{:.2f}'.format(
            sum(float(v) for v in self.amount_details.values())
        ))

    def to_dict(self):
        data = {
            'amount': {
                "total": "0.00"
            },
            "item_list": {
                "items": [],
            }
        }

        direct_fields = [
            'invoice_number',
            'description',
            'reference_id',
            'custom',
        ]
        for field in direct_fields:
            value = getattr(self, field, None)
            if value:
                data[field] = getattr(self, field)

        data['amount']['total'] = str(self.amount_total)
        data['amount']['currency'] = self.currency
        if self.amount_details:
            data['amount']['details'] = self.amount_details

        data['item_list']['items'] = self.items

        return data

    @classmethod
    def on_post_save(cls, instance: 'PaypalTransaction', raw=None, **kwargs):
        if raw:  # 导入fixture时不运行
            return

        if instance.tracker.has_changed('state') and instance.state == PaypalTransaction.STATE_APPROVED:
            payment_succeed.send(PaypalTransaction, instance=instance)


post_save.connect(PaypalTransaction.on_post_save, PaypalTransaction)



from django.test import RequestFactory, TestCase

from paypal.models import PaypalTransaction
from paypal.views import PaypalRedirectView


class PaypalRedirectViewTest(TestCase):
    def test_create(self):
        tx = PaypalTransaction(description="hello")
        tx.add_item("vpn", 'USD', 1.11)
        tx.save()

        view = PaypalRedirectView.as_view()
        factory = RequestFactory()
        req = factory.get('/')

        resp = view(req, pk=tx.pk)
        self.assertTrue('paypal' in resp.url, '跳转到paypal')


"""
原则上需要定义三个view name：
    - paypal_redirect
    - paypal_return
    - paypal_cancel

Example urlpatterns
-------------------

urlpatterns = [
    url('^redirect/(?P<pk>.*?)/$', PaypalRedirectView.as_view(), name='paypal_redirect'),
    url('^return/$', PaypalReturnView.as_view(), name='paypal_return'),
    url('^cancel/$', PaypalCancelView.as_view(), name='paypal_cancel'),
]
"""

import paypalrestsdk
from django.conf import settings
from django.http import HttpResponseBadRequest, HttpResponseServerError
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views import View
from django.views.generic import TemplateView

from paypal.models import PaypalTransaction


def _get_api():
    conf = settings.PAYPAL.copy()
    if conf.pop('sandbox', False):
        conf['mode'] = 'sandbox'
    else:
        conf['mode'] = 'live'
    return paypalrestsdk.Api(conf)


def example_paypal_create_view(request):
    """  创建transaction的样例 """
    tx = PaypalTransaction(description="hello")
    tx.add_item("vpn", 'USD', 1.11)
    tx.save()

    return redirect('paypal_redirect', pk=tx.pk)


class PaypalRedirectView(View):
    """
    他人生成PaypalTransaction以后，跳转到本view就能重定向到paypal，继续完成支付流程
    样例见： example_paypal_create_view
    """

    def get_return_url(self):
        """ 定义从paypal支付成功以后返回的URL，按需定义 """
        return settings.PAYPAL['SERVER_URL'] + reverse('paypal_return')

    def get_cancel_url(self):
        """ 定义用户取消paypal支付时，返回的页面URL，按需定义 """
        return settings.PAYPAL['SERVER_URL'] + reverse('paypal_cancel')

    def handle_redirect(self, redirect_url):
        """ 得到paypal支付地址以后，定义如何跳转（默认为直接跳转，可以自己改成打开loading页面，用JS跳转 """
        return redirect(redirect_url)

    def payment_create_error(self, payment):
        """ 创建支付失败时的response，按需定义 """
        print("handle_payment_create_error", payment.error)
        return HttpResponseServerError()

    def get(self, request, pk):
        tx = get_object_or_404(PaypalTransaction, pk=pk, state=PaypalTransaction.STATE_CREATED)

        if tx.payment_id:
            payment = paypalrestsdk.Payment.find(tx.payment_id, api=_get_api())
        else:
            params = self.make_params(tx)
            payment = paypalrestsdk.Payment(params, api=_get_api())
            ok = payment.create()
            if not ok:
                return self.payment_create_error(payment)

            tx.payment_id = payment.id
            tx.save()

        for link in payment.links:
            if link.method == 'REDIRECT':
                return self.handle_redirect(link.href)

    def make_params(self, tx):
        params = {
            'intent': 'sale',
            'payer': {
                'payment_method': 'paypal',
            },
            "transactions": [
                tx.to_dict(),
            ],
            "redirect_urls": {
                "return_url": self.get_return_url(),
                "cancel_url": self.get_cancel_url(),
            }
        }

        if tx.note_to_payer:
            params['note_to_payer'] = tx.note_to_payer

        return params


class PaypalReturnView(View):
    """
    从paypal支付成功以后，回来的页面，
    页面自己会处理transaction成功的逻辑，成功以后会发送 payment_succeed signal
    业务逻辑只需要connect payment_succeed，然后做相应的处理即可
    """

    def payment_succeed(self, tx):
        """ 支付成功以后的返回结果，按需定义 """
        return render(self.request, 'paypal/return.html', context=dict(transaction=tx))

    def payment_execute_error(self, payment):
        """ 执行支付失败后的返回结果，按需定义 """
        print("handle_payment_execute_error", payment.error)
        return HttpResponseServerError()

    def get(self, request):
        payment_id = request.GET.get('paymentId')
        payer_id = request.GET.get('PayerID')

        if not payment_id or not payer_id:
            return HttpResponseBadRequest()

        tx = get_object_or_404(PaypalTransaction, payment_id=payment_id)
        if tx.state == tx.STATE_APPROVED:
            return self.payment_succeed(tx)

        payment = paypalrestsdk.Payment.find(payment_id, api=_get_api())
        if payment.state != tx.STATE_APPROVED:
            ok = payment.execute({'payer_id': payer_id})
            if not ok:
                return self.payment_execute_error(payment)

        tx.state = payment.state
        tx.payer_id = payment.payer.payer_info.payer_id
        tx.payer_email = payment.payer.payer_info.email
        tx.payer_data = payment.payer.to_dict()

        tx.payee_merchant_id = payment.transactions[0].payee.merchant_id
        tx.save()  # post_save 时，如果初次成功，会抛 payment_succeed signal

        return self.payment_succeed(tx)


class PaypalCancelView(TemplateView):
    template_name = 'paypal/cancel.html'
